/* 
 * File:   Box.h
 * Author: Kolja Junginger
 *
 * Created on February 26, 2020, 5:05 PM
 * 
 * For more details on the algorithms see 
 * "Certified Approximation Algorithms for the Fermat Point" by 
 * Kolja Junginger, Ioannis Mantas, Evanthia Papadopoulou, Martin Suderland, and Chee Yap:
 * http://www1.pub.informatik.uni-wuerzburg.de/eurocg2020/data/uploads/papers/eurocg20_paper_72.pdf
 */

#ifndef BOX_H
#define BOX_H

#include "Interval.h"
#include "Point.h"
#include<vector>

class Box {
public:
    Box(Interval P, Interval Q);
    Box();
    Interval Ix;
    Interval Iy;
    bool isempty;
    
    double width() const {
        return Ix.max - Ix.min;
    }
    double height() const {
        return Iy.max - Iy.min;
    }
    double radius() const {
        return sqrt( (pow(width(),2) + pow(height(),2)) / 2 );
    }
    Point center() const {
        Point C = Point((Ix.max + Ix.min)/2, (Iy.max + Iy.min)/2);
        return C;
    }
    std::vector<Point> corners() const {
        Point L_min = Point(Ix.min,Iy.min);
        Point L_max = Point(Ix.min,Iy.max);
        Point R_min = Point(Ix.max,Iy.min);
        Point R_max = Point(Ix.max,Iy.max);
        std::vector<Point> Corners= {L_min, L_max, R_min,R_max};
        return Corners;
    }
    std::vector<Box> split(){
        Point C = center();
        Box B1 = Box( Interval(Ix.min,C.x),Interval(Iy.min,C.y) ); 
        Box B2 = Box( Interval(C.x,Ix.max),Interval(Iy.min,C.y) ); 
        Box B3 = Box( Interval(Ix.min,C.x),Interval(C.y,Iy.max) ); 
        Box B4 = Box( Interval(C.x,Ix.max),Interval(C.y,Iy.max) );    
        
        std::vector<Box> children = {B1,B2,B3,B4};
        
        return children;
    }
};

inline Box::Box(Interval P, Interval Q)
{
    Ix = P;
    Iy = Q;
    if (Ix.isempty || Iy.isempty){
        isempty = true;
        //            std::cout<<"WARNING: You have defined an empty box."<<std::endl;
    }
    else {
        isempty = false;
    }
}

inline Box::Box()
{
    Ix = Interval(0,0);
    Iy = Interval(0,0);
    isempty = false;
}

inline std::ostream& operator<<(std::ostream& os, Box& B) {
    return os << "I.x: " << B.Ix << "I.y: "<< B.Iy << "\n";
}


inline bool inclusion(Box& B1, Box& B2){
    //     Checks whether Box B1 is a subset of B2,
    //    !!!Caution!!! Different from Martin's matlab implementation!!! Inputs swapped!
    bool answer = (subset(B1.Ix,B2.Ix) && subset(B1.Iy,B2.Iy));
    return answer;
}

inline bool exclusion(Box& B1,Box& B2){
    // Returns true if intersection of Boxes B1 and B2 is empty
    
    Interval Intx = intersection(B1.Ix,B2.Ix);
    Interval Inty = intersection(B1.Iy,B2.Iy);
    
    return Intx.isempty || Inty.isempty;
}

inline Box intersection(Box& B1,Box& B2){
    // Returns the intersection of Boxes B1 and B2 
    
    Interval Intx = intersection(B1.Ix,B2.Ix);
    Interval Inty = intersection(B1.Iy,B2.Iy);
    Box Bint = Box(Intx,Inty);
    
    return Bint;
}

inline bool contained(const Point& P,const Box& B){
    // Returns true if Point P is contained in Box B
    
    bool answer = (subset(Interval(P.x,P.x),B.Ix) && subset(Interval(P.y,P.y),B.Iy));            
    
    return answer;
}

inline Box center2box(const Point& center, double width){
    Interval B_x = Interval(center.x - width/2, center.x + width/2);
    Interval B_y = Interval(center.y - width/2, center.y + width/2);
    return Box(B_x,B_y);
}

inline Box scaleBox(const Box& B, double c){
    // Computes Box cB = c*B where the dilated box has its center at the midpoint of the box B
    
    Point Center = B.center();
    double new_width = c * B.width();
    //double new_height = c * B.height();
    Box cB = center2box(Center,new_width);
    return cB;
}

inline Interval normB(const Box& B) {
    // Computes the set {norm(p) | p in B} as an interval.
    std::vector<Point> Cor = B.corners();
    std::vector<double> norm_corners;
    for (int i=0;i<Cor.size();i++){
        norm_corners.push_back(normP(Cor[i]));                
    }
    double min_norm = * std::min_element(norm_corners.begin(),norm_corners.end());
    double max_norm = * std::max_element(norm_corners.begin(),norm_corners.end());
    
    bool testy = (B.Ix.min <= 0 && 0 <= B.Ix.max); //box intersects the y-axis
    bool testx = (B.Iy.min <= 0 && 0 <= B.Iy.max); // box intersects the x-axis  
    Interval I;
    
    if (testy && testx) {
        I = Interval(0,max_norm);
    }
    else if (testy && !testx) {
        I = Interval( std::min(abs(B.Iy.min),abs(B.Iy.max)), max_norm);
    }
    else if (!testy && testx) {
        I = Interval( std::min(abs(B.Ix.min),abs(B.Ix.max)), max_norm);
    }
    else {
        I = Interval(min_norm, max_norm);
    }
    return I;
}


inline Interval sinI(const Box& B) {
    // For any point p = (px, py) , we define sin(p) := px/||p|| and cos(p) := py/||p||.
    
    std::vector<Point> Cor = B.corners();    // corners: {L_min, L_max, R_min,R_max}
    double min_sin = 1;
    double max_sin = -1;
    for (int i=0;i<Cor.size();i++){
        double Sini = Cor[i].y / normP(Cor[i]);
        min_sin = std::min(min_sin, Sini);
        max_sin = std::max(max_sin, Sini);
    }
    
    // Test whether the origin is below or above of box B:
    Interval I;
    if (B.Ix.min <=0 && 0 <= B.Ix.max){
        if (0 < B.Iy.min){ // origin is below box B
            I = Interval(min_sin,1);
        }
        else if (B.Iy.max < 0){ // origin is above of box B
            I =  Interval(-1,max_sin);
        }
        else // origin is in box B
            I = Interval(-1,1);
    }
    else // origin is neither below nor above of box B (also not contained in B)
        I = Interval(min_sin,max_sin);
    
    return I;
}

inline Interval cosI(const Box& B) {
    // For any point p = (px, py) , we define sin(p) := px/||p|| and cos(p) := py/||p||.
    
    std::vector<Point> Cor = B.corners();    // corners: {L_min, L_max, R_min,R_max}
    
    double Cosi;
    double min_cos = 1;
    double max_cos = -1;
    for (int i=0;i<Cor.size();i++){
        Cosi = Cor[i].x / normP(Cor[i]);
        min_cos = std::min(min_cos, Cosi);
        max_cos = std::max(max_cos, Cosi);
    }
    
    // Test whether the origin is left or right of box B
    Interval I;
    if (B.Iy.min <=0 && 0 <= B.Iy.max){
        if (0 < B.Ix.min){ // origin is left of box B
            I = Interval(min_cos,1);
        }
        else if (B.Ix.max < 0){ // origin is right of box B
            I = Interval(-1,max_cos);
        }
        else // origin is in box B
            I = Interval(-1,1);
    }
    else // origin is neither left nor right of box B (also not contained in B)
        I = Interval(min_cos,max_cos);
    
    return I;
}

inline Interval powI(Interval I, int n) {
    double min_to_n = pow(I.min,n);
    double max_to_n = pow(I.max,n);
    double Min;
    double Max;
    if (fmod(n,2) ==0) {
        Min = min_to_n;
        Max = max_to_n;
    }
    else if (I.min<0 && 0 < I.max){
        Min = std::min(min_to_n, max_to_n);
        Max = std::max(min_to_n, max_to_n);
    }
    else {
        Min = 0;
        Max = std::max(min_to_n, max_to_n);
    }
    
    return Interval(Min,Max);
}




#endif /* BOX_H */

