/* 
 * File:   Interval.h
 * Author: Kolja Junginger
 *
 * Created on February 26, 2020, 4:53 PM 
 * 
 * For more details on the algorithms see 
 * "Certified Approximation Algorithms for the Fermat Point" by 
 * Kolja Junginger, Ioannis Mantas, Evanthia Papadopoulou, Martin Suderland, and Chee Yap:
 * http://www1.pub.informatik.uni-wuerzburg.de/eurocg2020/data/uploads/papers/eurocg20_paper_72.pdf
 */

#ifndef INTERVAL_H
#define INTERVAL_H

#include<cmath>
#include<vector>
#include<iostream>


class Interval {
public:
    Interval(double a, double b);
    Interval();
    double min;
    double max;
    bool isempty = false;
    
    double length() const {
        double l;
        if (!isempty){
            l = max - min;
        }
        else {
            l = std::nan("1");
        }
        return l;
    }
};

inline Interval::Interval(double a, double b)
{
    min = a;
    max = b;
    if (b < a) {
        isempty = true;
        //            std::cout<<"Warning, defined empty interval: "<<'('<<a<<','<<b<<')'<<std::endl;
    }
    else
        isempty = false;
}

inline Interval::Interval()
{
    min  = 0;
    max = 0;
    isempty = false;
}

inline std::ostream& operator<<(std::ostream& os, Interval& I) {
    return os << '(' << I.min << ',' << I.max << ")\n";
}

inline bool operator==(const Interval& I, const Interval& J) {
    return (I.min == J.min && I.max == J.max);
}

inline bool operator!=(const Interval& I, const Interval& J) {
    return !(I == J);
}

inline Interval operator+(const Interval& I, const Interval& J) {
    Interval sumIJ = Interval(I.min + J.min, I.max + J.max);
    return sumIJ;
}

inline Interval operator-(const Interval& I, const Interval& J) {
    Interval subIJ = Interval(I.min - J.max, I.max - J.min);
    return subIJ;
}

inline Interval operator*(const Interval& I, const Interval& J) {
    std::vector<double> options = {I.min*J.min, I.min*J.max, I.max*J.min, I.max*J.max};
    double  min = * std::min_element(options.begin(),options.end());
    double  max = * std::max_element(options.begin(),options.end());
    Interval mul = Interval(min, max);
    return mul;
}

inline Interval operator/(const Interval& I, const Interval& J) {
    
    double inf = std::numeric_limits<double>::infinity();
    
    Interval Jinverse;
    if (0 < J.min || J.max < 0){
        Jinverse = Interval(1.0/J.max,1.0/J.min);
    }
    else if (J.min == 0){
        Jinverse = Interval(1.0/J.max, inf);
    }
    else if (J.max == 0){
        Jinverse = Interval(-inf,1.0/J.min);
    }
    else
        Jinverse = Interval(-inf,inf);
    
    Interval div = I*Jinverse;
    return div;
}

inline bool subset(const Interval& I,const Interval& J) {
    // checks whether Interval I is a subset of J. 
    
    bool answer = ( (J.min <= I.min) && (I.max <= J.max) );
    return answer;
}

inline Interval intersection(const Interval& I,const Interval& J) {
    // returns the intersection of Intervals I and J. 
    
    Interval Int = Interval( std::max(I.min,J.min), std::min(I.max,J.max) );
    
    if (Int.min  > Int.max){
        Int.isempty = true;
    }
    
    return Int;
}

inline Interval convexhull(const Interval& I,const Interval& J) {
    return Interval(std::min(I.min,J.min), std::max(I.max,J.max) );
}

#endif /* INTERVAL_H */

