/* 
 * File:   boxsubdivison.cpp
 * Author: Kolja Junginger
 * 
 * For more details on the algorithms see 
 * "Certified Approximation Algorithms for the Fermat Point" by 
 * Kolja Junginger, Ioannis Mantas, Evanthia Papadopoulou, Martin Suderland, and Chee Yap:
 * http://www1.pub.informatik.uni-wuerzburg.de/eurocg2020/data/uploads/papers/eurocg20_paper_72.pdf
 */

#include "predicates.h"

using namespace std;

inline Box boundingbox(const vector<Point>& foci);
inline Box boundingbox(queue<Box>& Q);

inline Point fermatpoint(const vector<Point>& foci, double epsilon){
    /* Given a foci set of k points, the Fermat point FP is the point in the plane
     * that minimizes the sum of the distances to the foci. The algorithm fermatpoint
     * computes an epsilon-approximation of the Fermat point FP, that is, a point FP'
     * satisfying |FP'− FP| < epsilon. The algorithm is based on the subdivision 
     * paradigm, which we combine with the Newton method used for speed-ups and 
     * certification it is certified in the sense of interval methods.   
     * We assume that the Fermat point is unique and not equal to a focus. 
     * 
     * We keep on splitting boxes using different kinds of predicates while we 
     * exclude boxes that are guaranteed not to contain FP and we keep boxes 
     * that might contain FP. We test whether we can already approximate FP well
     * enough by putting a bounding box around all boxes, which we have not 
     * excluded yet. Further, we use the Newton box operator N, which has the 
     * following properties: 
     * 
     * (1) If N(B) is a subset of B, then FP is in N(B).
     * (2) If FP is in B, then FP is in N(B). 
     * (3) If the intersection of N(B) and B is empty,then FP is not in B.
     * 
     * We perform the Newton test (1) for each box and in case of success we 
     * continue the subdivision only with N(B) and its (grand)children.
     */
    
    cout<<" --- Starting box subdivision --- "<<endl;
    cout<<"Epsilon = "<<epsilon<<endl;
    
    clock_t begin = clock();
    
    Box B0 = boundingbox(foci);
    
    queue<Box> Q;
    Q.push(B0);
    
    while(!CEpsilon(Q,epsilon)){
        
        if (Q.size()==0){
            cout<<"Warning, Size: "<<Q.size()<<endl;   
        }
        
        Box B = Q.front();
        Q.pop();
        
        
        if (CGradient(B,foci)){ // If Fermat Point might be contained in Box B
            Box B2 = scaleBox(B,2);
            Box NB2 = Newton(B2,foci);
            
            if (inclusion(NB2,B2)){ // If Newton box is contained in 2B.
                cout<<"NEWTON test succeeded."<<endl;
                
                while(!Q.empty()){ // Empty the queue Q:
                    Q.pop();
                }
                vector<Box> children = NB2.split();                
                for(int i=0;i < children.size();i++){
                    vector<Box> grandchildren = children[i].split();
                    for(int j=0;j < grandchildren.size();j++){
                        Q.push(grandchildren[j]);
                    }
                }
            }
            else { // " that means, NEWTON test NOT succeeded."
                Box NB = intersection(NB2,B);
                if (!NB.isempty){
                    vector<Box> children = NB.split();
                    for(int i=0;i < children.size();i++){
                        Q.push(children[i]);
                    }
                }
            }
        }
    }
    cout<<" --- end of box subdivision ---  "<<endl;
    
    Box BB = boundingbox(Q);
    Point fp = BB.center();
    
    print_time(begin);
    
    int prec = precision(epsilon, fp.x);
    streamsize default_prec = cout.precision();
    cout << "The approximated Fermat point with boxsubdivision is " << setprecision(prec) << fp << endl; 
    cout.precision(default_prec);
    
    return fp;
}


Box boundingbox(const vector<Point>& foci){
    /* This function computes a bounding box for the set of Points foci. */
    
    double xmin = foci[0].x;
    double xmax = foci[0].x;
    double ymin = foci[0].y;
    double ymax = foci[0].y;
    
    for(int i=1;i<foci.size();i++){
        xmin  = min(foci[i].x,xmin); 
        xmax  = max(foci[i].x,xmax); 
        ymin  = min(foci[i].y,ymin); 
        ymax  = max(foci[i].y,ymax); 
    }
    
    Box B0 = Box(Interval(xmin,xmax),Interval(ymin,ymax));
    
    return B0;
}

Box boundingbox(queue<Box>& Q){
    /* This function computes a bounding box for the queue of Boxes Q and 
     it empties the queue Q. */
    
    Box B = Q.front();
    Q.pop();
    Interval Ix = B.Ix;
    Interval Iy = B.Iy;
    
    while(!Q.empty()){
        Box B = Q.front();
        Q.pop();
        Ix = convexhull(Ix, B.Ix);
        Iy = convexhull(Iy, B.Iy);
    }
    Box BB = Box(Ix,Iy);
    
    return BB;
}
