/* 
 * File:   derivatives.h
 * Author: Kolja Junginger
 *
 * Created on March 3, 2020, 9:16 AM
 * 
 * * For more details on the algorithms see 
 * "Certified Approximation Algorithms for the Fermat Point" by 
 * Kolja Junginger, Ioannis Mantas, Evanthia Papadopoulou, Martin Suderland, and Chee Yap:
 * http://www1.pub.informatik.uni-wuerzburg.de/eurocg2020/data/uploads/papers/eurocg20_paper_72.pdf
 */

#ifndef DERIVATIVES_H
#define DERIVATIVES_H

#include "Point.h"
#include "Interval.h"
#include "Box.h"

using namespace std;

Interval delx(const Box& B, const vector<Point>& foci);
Interval dely(const Box& B, const vector<Point>& foci);
Interval delx2(const Box& B, const vector<Point>& foci);
Interval dely2(const Box& B, const vector<Point>& foci);
Interval delxdely(const Box& B, const vector<Point>& foci);
vector<vector<Interval>> hessian(const Box& B, const vector<Point>& foci);
vector<Interval> jaccobian(const Box& B, const vector<Point>& foci);
Point koperator(const Point & P, const vector<Point> & foci);
Box Newton(const Box& B,const vector<Point>& foci);
vector<vector<Interval>> InverseMatrix(const vector<vector<Interval>>& M);
vector<Interval> operator*(const vector<vector<Interval>>& M1, const vector<Interval>& M2);
vector<vector<Interval>> operator-(const vector<vector<Interval>>& M1, const vector<vector<Interval>>& M2);
vector<Interval> operator-(const vector<Interval>& M1, const vector<Interval>& M2);

inline Box Newton(const Box& B,const vector<Point>& foci){
    /* This function Newton maps a Box B to a new box N(B) using the Newton formula by Moore and Nickel:
     MOORE, R. E. : Interval Analysis, Englewood Cliffs, New Jersey: Prentice Hall, Inc. 1966.
     NICKEL, K.   : TRIPLEX-ALGOL and Applications. Interner Bericht des Lehrstuhls fuer Numerisehe Mathematik und Grossrechenanlagen der Universitaet (TH) Karlsruhe.
     N(B) = m_B - J^-1 (B) * f(m_B)
     */
    
    int k = foci.size();
    
    double cx = (B.Ix.min + B.Ix.max)/2;
    double cy = (B.Iy.min + B.Iy.max)/2;
    Box center = Box(Interval(cx,cx),Interval(cy,cy)); // m_B
    vector<Interval> Cent = {Interval(cx,cx),Interval(cy,cy)};
    
    vector<Interval> gradient_phi           = jaccobian(center,foci);
    vector<vector<Interval>> hessianInverse = InverseMatrix(hessian(B,foci));
    vector<Interval> Newton_B               = Cent - (hessianInverse * gradient_phi);
    
    Box Bout = Box(Newton_B[0],Newton_B[1]);
    return Bout ;
}



inline vector<Interval> jaccobian(const Box& B, const vector<Point>& foci){
    // computes the 1st derivatives (gradient) at box B of phi - the Fermat distance function. -> Function f in RootNewton.m
    Interval derivative_x = delx(B,foci);
    Interval derivative_y = dely(B,foci);
    vector<Interval> J = {derivative_x,derivative_y};
    return J;
}

inline vector<vector<Interval>> hessian(const Box& B, const vector<Point>& foci){
    // computes the 2nd derivatives (Hessian) of phi - the Fermat distance function. 
    Interval dx2  = delx2(B,foci);
    Interval dy2  = dely2(B,foci);
    Interval dxdy = delxdely(B,foci);
    vector<vector<Interval>> J = {{dx2,dxdy},{dxdy,dy2}};    
    return J;
}


inline Interval delx(const Box& B, const vector<Point>& foci){
    // computes the derivative of the Fermat distance function in x direction
    Interval dx = Interval(0,0);
    int k = foci.size();
    
    for (int i=0;i<k;i++){
        Interval B_shifted_x = Interval(B.Ix.min - foci[i].x, B.Ix.max - foci[i].x);
        Interval B_shifted_y = Interval(B.Iy.min - foci[i].y, B.Iy.max - foci[i].y);
        Box B_shifted = Box(B_shifted_x,B_shifted_y);
        Interval cossh = cosI(B_shifted);
        dx = dx + cossh;
    }
    
    return dx;
}


inline Interval dely(const Box& B, const vector<Point>& foci){
    // computes the derivative of the Fermat distance function in x direction
    Interval dy = Interval(0,0);
    int k = foci.size();
    
    for (int i=0;i<k;i++){
        Interval B_shifted_x = Interval(B.Ix.min - foci[i].x, B.Ix.max - foci[i].x);
        Interval B_shifted_y = Interval(B.Iy.min - foci[i].y, B.Iy.max - foci[i].y);
        Box B_shifted = Box(B_shifted_x,B_shifted_y);
        Interval sinsh = sinI(B_shifted);
        dy = dy + sinsh;
    }
    
    return dy;
}

inline Interval delx2(const Box& B, const vector<Point>& foci){
    // computes the 2nd derivative of the Fermat distance function in x direction
    Interval dx2 = Interval(0,0);
    int k = foci.size();
    for (int i=0;i<k;i++){
        Interval B_shifted_x = Interval(B.Ix.min - foci[i].x, B.Ix.max - foci[i].x);
        Interval B_shifted_y = Interval(B.Iy.min - foci[i].y, B.Iy.max - foci[i].y);
        Box B_shifted = Box(B_shifted_x,B_shifted_y);
        Interval sI = sinI(B_shifted);
        Interval numerator = sI * sI;
        Interval denominator = normB(B_shifted);
        Interval fraction = numerator / denominator;
        dx2 = dx2 + fraction;
    }
    
    return dx2;
}

inline Interval dely2(const Box& B, const vector<Point>& foci){
    // computes the 2nd derivative of the Fermat distance function in y direction
    Interval dy2 = Interval(0,0);
    int k = foci.size();
    for (int i=0;i<k;i++){
        Interval B_shifted_x = Interval(B.Ix.min - foci[i].x, B.Ix.max - foci[i].x);
        Interval B_shifted_y = Interval(B.Iy.min - foci[i].y, B.Iy.max - foci[i].y);
        Box B_shifted = Box(B_shifted_x,B_shifted_y);
        Interval cosinus = cosI(B_shifted);
        Interval numerator = cosinus * cosinus;
        Interval denominator = normB(B_shifted);
        Interval fraction = numerator / denominator;
        dy2 = dy2 + fraction;
    } 
    return dy2;
}

inline Interval delxdely(const Box& B, const vector<Point>& foci){
    Interval dxdy = Interval(0,0);
    int k = foci.size();
    for (int i=0;i<k;i++){
        Interval B_shifted_x = Interval(B.Ix.min - foci[i].x, B.Ix.max - foci[i].x);
        Interval B_shifted_y = Interval(B.Iy.min - foci[i].y, B.Iy.max - foci[i].y);
        Box B_shifted = Box(B_shifted_x,B_shifted_y);
        Interval sI = sinI(B_shifted);
        Interval cosinus = cosI(B_shifted);
        Interval numerator = sI * cosinus;
        Interval denominator = normB(B_shifted);
        Interval fraction = numerator / denominator;
        dxdy = dxdy - fraction;
    }
    
    return dxdy;
}


/* Matrix operations */


inline vector<Interval> operator*(const vector<vector<Interval>>& M1, const vector<Interval>& M2) {
    // Computes Bout = M1*M2 a 2x2 matrix M1 and 2x1 matrix M2.
    
    int rows1       = M1.size();
    int columns1    = M1[0].size();
    int rows2       = M2.size();
    
    vector<int> s1 = {rows1,columns1}; // rows, columns
    
    if ((s1[1] !=rows2 ) || (s1[0]!=2 && s1[1] !=2) || (rows2 !=2))  {
        throw runtime_error("ERROR: Matrices do not have the fitting size for multiplication.");
    }

    Interval Mout1 = (M1[0][0] * M2[0] + M1[0][1] * M2[1]);
    Interval Mout2 = (M1[1][0] * M2[0] + M1[1][1] * M2[1]);        
   
    vector<Interval> Mout = {  Mout1,  Mout2};
    
    return Mout;
}


inline vector<vector<Interval>> operator-(const vector<vector<Interval>>& M1, const vector<vector<Interval>>& M2) {
    // Computes Bout = B1*B2 for two Boxes, interpreted as matrix multiplication.
    
    int rows1       = M1.size();
    int columns1    = M1[0].size();
    int rows2       = M2.size();
    int columns2    = M2[0].size();
    
    vector<int> s1 = {rows1,columns1}; // rows, columns
    vector<int> s2 = {rows2,columns2}; // rows, columns
    
    if ( (s1[0] !=s2[0]) || (s1[1] !=s2[1])  ) {
        throw runtime_error("ERROR: Matrices do not have the fitting size for multiplication.");
    }
    
    vector<vector<Interval>> Mout = {{Interval(),Interval()},{Interval(),Interval()}};
    
    for (int i=0; i < s1[0]; i++){
        for (int j=0; j < s1[1]; j++){
            Interval M1ij = M1[i][j];
            Interval M2ij = M2[i][j];
            Mout[i][j] = M1ij -  M2ij;
        }
    }
    
    return Mout;
}


inline vector<Interval> operator-(const vector<Interval>& M1, const vector<Interval>& M2) {
    // Computes Bout = B1-B2 for two intervals
    
    int rows1       = M1.size();
    int rows2       = M2.size();
    
    
    if ( rows1!=rows2   ) {
        throw runtime_error("ERROR: Matrices do not have the fitting size for subtraction.");
    }
    
    vector<Interval> Mout = {Interval(),Interval()};
    
    for (int i=0; i < rows1; i++){
        Interval M1i = M1[i];
        Interval M2i = M2[i];
        Mout[i] = M1i -  M2i;
    }
    
    return Mout;
}


inline vector<vector<Interval>> InverseMatrix(const vector<vector<Interval>>& M) {
    // Computes the inverse M^-1 of a square sized interval matrix M
    
    int rows       = M.size();
    int columns    = M[0].size();
    vector<int> size = {rows,columns}; 
    
    if (size[1] !=size[0]) {
        throw runtime_error("ERROR: Matrix is not a square matrix.");
    }
    
    Interval det =  (M[0][0] * M[1][1]) - (M[0][1] * M[1][0]);   
    Interval zero = Interval(0,0);
    
    Interval M00 =  M[1][1] / det;
    Interval M01 = (zero - M[0][1]) / det;
    Interval M10 = (zero - M[1][0]) / det;
    Interval M11 =  M[0][0] / det;
    
    vector<vector<Interval>> Mout = { {M00,M01} , {M10,M11} };
    return Mout;
}


#endif /* DERIVATIVES_H */

