/* 
 * File:   focifermatpointtest.cpp
 * Author: Kolja Junginger
 * 
 * For more details on the algorithms see 
 * "Certified Approximation Algorithms for the Fermat Point" by 
 * Kolja Junginger, Ioannis Mantas, Evanthia Papadopoulou, Martin Suderland, and Chee Yap:
 * http://www1.pub.informatik.uni-wuerzburg.de/eurocg2020/data/uploads/papers/eurocg20_paper_72.pdf
 */

#include "Point.h"
#include "Box.h"
#include "derivatives.h"

using namespace std;


inline vector<Point> foci_minus_one(vector<Point> foci, int index){
    foci.erase(foci.begin() + index);
    return foci;
}

inline int focusisfermatpoint(const vector<Point>& foci){
    /* This function focusisfermatpoint a simple
     * O(k^2d) preprocessing algorithm to verify if the Fermat point is equal 
     * to one of the foci.
     * A focus a of foci set A is the Fermat point FP of A if and only if 
     * |Gradient phi_{A-a}(a)| < 1. For each focus a, it 
     * tests, whether the length of the gradient of the other foci A-a is at 
     * most 1 at one of the foci. If yes, then we output the focus index if no, 
     * then we output -1.
     */
    
    clock_t begin = clock();
    
    int k = foci.size();
    int fermatpointind = -1;
    
    int focusind = 0;
    
    while ( (focusind < k) && (fermatpointind == -1) ){
        
        Point focus = foci[focusind];
        Box B = Box(Interval(focus.x,focus.x),Interval(focus.y,focus.y));
        vector<Point> foci_new =  foci_minus_one(foci,focusind);
        vector<Interval> gradient = jaccobian(B, foci_new);
        double length_gradient = sqrt( pow(gradient[0].min,2) + pow(gradient[1].min,2) );
        
        if (length_gradient <= 1) {
            fermatpointind = focusind;
        }
        
        focusind++;
    }
    
    if (fermatpointind == -1){
        cout<<"None of the foci equals the Fermat point."<<endl;
    }
    else {
        Point fp = foci[fermatpointind];
        cout<<"One of the foci equals the Fermat point: "<<fp;
    }
    
    print_time(begin);
    
    return fermatpointind;
}

