/* 
 * File:   predicates.h
 * Author: Kolja Junginger
 *
 * Created on March 3, 2020, 9:21 AM
 * 
 * For more details on the algorithms see 
 * "Certified Approximation Algorithms for the Fermat Point" by 
 * Kolja Junginger, Ioannis Mantas, Evanthia Papadopoulou, Martin Suderland, and Chee Yap:
 * http://www1.pub.informatik.uni-wuerzburg.de/eurocg2020/data/uploads/papers/eurocg20_paper_72.pdf
 */

#ifndef PREDICATES_H
#define PREDICATES_H


#include "derivatives.h"
#include<queue>

using namespace std;

inline bool CGradient(const Box& B,const vector<Point>& foci){
    /* This function returns true if 0 is contained in the gradient of Box B. */
    
    vector<Interval> gradient = jaccobian(B,foci);
    Interval zero = Interval(0,0);
    bool answer = (subset(zero,gradient[0]) && subset(zero,gradient[1]));            
    
    return answer;
    
}

inline bool CNewton(const Box& B,const vector<Point>& foci){
    /* This function returns true if the Newton box N(2B) is contained in the Box 2B. */
    
    Box B2 = scaleBox(B,2);
    Box NB2 = Newton(B2,foci);
    
    return inclusion(NB2,B2);            
    
}

inline bool CEpsilon(queue<Box> Blist, double epsilon){
    /* This function returns true if all boxes in Blist are contained in a bounding box having a radius smaller than epsilon. */
    
    if (Blist.empty()){
        throw runtime_error("The queue Q for the CEpsilon predicate is empty.");
    }
    Box B = Blist.front();
    Blist.pop();
    Interval Ix = B.Ix;
    Interval Iy = B.Iy;
    
    while(!Blist.empty()){
        Box B = Blist.front();
        Blist.pop();
        Ix = convexhull(Ix, B.Ix);
        Iy = convexhull(Iy, B.Iy);
    }
    
    double r = sqrt(pow(Ix.length(),2) + pow(Iy.length(),2) );
    bool answer = (r <= epsilon);
    return answer;
    
}

#endif /* PREDICATES_H */

