/* 
 * File:   weisfeld.cpp
 * Author: Kolja Junginger
 * 
 * For more details on these algorithms see 
 * "Certified Approximation Algorithms for the Fermat Point" by 
 * Kolja Junginger, Ioannis Mantas, Evanthia Papadopoulou, Martin Suderland, and Chee Yap:
 * http://www1.pub.informatik.uni-wuerzburg.de/eurocg2020/data/uploads/papers/eurocg20_paper_72.pdf
 */


#include "Point.h"
#include "Box.h"
#include "derivatives.h"

using namespace std;

Point center_of_mass(const vector<Point> & foci);
Point weisfeld(const vector<Point> & foci, double epsilon);
Point weisfeld_without_guarantee(const vector<Point> & foci, double epsilon);
Point koperator(const Point & P, const vector<Point> & foci);

inline Point weisfeld(const vector<Point> & foci, double epsilon) {
    /*
     The function weisfeld implements an algorithm based on Weiszfeld/Kuhn's iterative
     method (1937) to compute a sequence of points that converges to the
     Fermat point of a given foci set. The Fermat
     point is the point in the plane that minimizes the sum of the distances to the foci.
     We augment this iterative point sequence
     scheme by adding a guarantee for the computation, turning it into an
     epsilon-approximation algorithm. In order to verify the quality of the
     current point, we use the Newton inclusion predicate. We define a small box B
     with the current point p as center and width at most epsilon and map it
     to the new box N(B) using the Newton operator. If the box N(B) is contained in B,
     then we know that the Fermat point lies in N(B), and we output it.
     Otherwise, we keep on performing Weiszfeld steps while
     adjusting the box sizes, which is necessary for the correctness and
     termination of the algorithm.
     */
    
    bool skip_test = false;
    
    cout<<" --- Starting Weiszfeld's point sequence --- "<<endl;
    
    /* We use the center of mass of the foci as the starting point, since 
     it is a 2-approximation of the Fermat radius, see Cohen et al. 2016 */
    Point start = center_of_mass(foci);
    cout <<"Start point is: " <<start;
    cout<<"Epsilon = "<<epsilon<<endl;
    Point approx_FP;
    double scalefactor = 10;
    double dilation = 1.0/2;//1.0/2;//0.5;
    
    Point p = start;
    double box_epsilon = epsilon;
    double min_eps     = box_epsilon;
    
    clock_t begin = clock();
    
    bool stop = false;
    int i = 0;
    int freq = 1;
    int nr_Newton_tests = 0;
    
    while (!stop) {
        
        if (!skip_test || (i % freq==0) ){ // Decrease the frequency of performing Newton tests the more steps have been done.
            if (skip_test) {
                nr_Newton_tests ++;
                cout<<"Newton test no: "<<nr_Newton_tests<<endl;
            }
            Box B = center2box(p,box_epsilon);
            Box Bdil = scaleBox(B, dilation);
            Box NB = Newton(B,foci);
            Box NBdil = Newton(Bdil,foci);
            
            if (inclusion(NB,B)){  // if Box NB is contained in Box B
                stop = true;
                cout<<"Inclusion test succeeded after "<<i<<" Weisfeld steps."<<endl;                
                p = NB.center(); // center of the Newton box is most likely a better approximation than p (but both are correct).
                int prec = precision(box_epsilon, p.x);
                streamsize default_prec = cout.precision();
                cout.precision(prec);
                cout << "The " <<box_epsilon<<"-approximation of the Fermat point with Weisfeld is " <<  p << endl;
                cout.precision(default_prec);
            }
            else if (exclusion(Bdil, NBdil)){ // if Boxes Bdil and NBdil are disjoint
                if (box_epsilon*scalefactor <= epsilon){
                    box_epsilon = box_epsilon*scalefactor;
                    cout<<"...inreased box size to "<<box_epsilon<<endl;
                }
            }
            else { // i.e., Box B is not contained in box N(B) and boxes are not disjoint
                box_epsilon = box_epsilon/scalefactor;
                min_eps = min(min_eps,box_epsilon);
                cout<<"...reduced box size to "<<box_epsilon<<endl;
            }
            freq = freq*2;
        }
        Point q = koperator(p, foci);
        p = q;
        i++;
    }
    
    print_time(begin);
    
    approx_FP = p;
    return approx_FP;
}

inline Point koperator(const Point & P, const vector<Point> & foci) {
    /*
     For a set of k foci = {foci_1,...,foci_k} in R^2 and
     for a point P, place the weights 1/|p-foci_1, ..., 1/|p-foci_k|
     at the foci. Then the this function kOperator maps P to
     the center of mass of the weighted foci, i.e.:
     kOperator(P) = ( sum_i (foci_i)/(|P-foci_i|) ) / ( sum_i 1/|P-foci_i| )
     From another perspective the method can also be seen as moving in the
     direction of the negative gradient at the current point P, where the step
     length depends on the length of the gradient [Kuhn 1973].
     */
    
    int k = foci.size();
    
    // Computation of Kuhn's formula (Weisfeld):
    
    double denominator=0;
    Point enumerator = Point(0,0);
    
    for (int i = 0; i < k; i++) {
        double d = dist(foci[i], P);
        denominator = denominator + 1.0/d;
        Point focus = foci[i];
        Point frac = (focus / d);
        enumerator = enumerator + frac;
    }
    
    Point mappedpoint = enumerator / denominator;
    
    return mappedpoint;
}


inline Point center_of_mass(const vector<Point> & foci) {
    
    Point center = Point(0, 0);
    int k = foci.size();
    for (int i = 0; i < k; i++) {
        Point focus = foci[i];
        center = center + focus;
    }
    double kk = k;
    
    Point masscenter = (center /kk);
    
    return masscenter;
}



inline Point weisfeld_without_guarantee(const vector<Point> & foci, double epsilon) {
    /*
     * This is Weiszfeld's point sequence scheme, which is stopped as soon as
     * two consecutive points have a distance smaller than epsilon. This naive
     * algorithm has no guarantee at all and can output points arbitrarily
     * far away from the real Fermat point.
     */
    
    cout<<" --- Starting Weiszfeld's point sequence --- "<<endl;
    
    int limit = pow(10, 6);
    Point start = center_of_mass(foci);
    cout <<"Start point is: " <<start;
    Point approx_FP;
    
    Point p = start;
    
    bool stop = false;
    int i = 0;
    while (!stop and i < limit) {
        Point q = koperator(p, foci);
        if (dist(p, q) < epsilon) {
            stop = true;
            cout << "Close to FP after " << i << " steps" << endl;
            cout << "FP is " << q << endl;
        }
        p = q;
        i++;
    }
    approx_FP = p;
    return approx_FP;
}
